"""Constant values for neosctl."""

import pathlib

PROFILE_FILENAME = "profile"
ENV_FILENAME = "env"
CREDENTIAL_FILENAME = "credential"
LOG_FILENAME = "error.log"
ROOT_FILEPATH = pathlib.Path().home() / ".neosctl"
PROFILE_FILEPATH = ROOT_FILEPATH / PROFILE_FILENAME
ENV_FILEPATH = ROOT_FILEPATH / ENV_FILENAME
CREDENTIAL_FILEPATH = ROOT_FILEPATH / CREDENTIAL_FILENAME
LOG_FILEPATH = ROOT_FILEPATH / LOG_FILENAME


DEFAULT_PROFILE = "default"


SUCCESS_CODE = 200
REDIRECT_CODE = 300
BAD_REQUEST_CODE = 400
UNAUTHORISED_CODE = 401

GATEWAY = "gateway"
REGISTRY = "registry"
IAM = "iam"
